# Generated by Django 3.2.5 on 2021-10-06 17:43

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('data', '0036_auto_20211006_1743'),
    ]

    operations = [
        migrations.AlterField(
            model_name='datatransformerconstants',
            name='id',
            field=models.UUIDField(blank=True, default=uuid.UUID('9a3515db-6b03-4be1-bd5e-949656b55c10'), primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='hardwarestatsdatatransformer',
            name='id',
            field=models.UUIDField(blank=True, default=uuid.UUID('23a04c6b-70d5-4825-be8d-038a6c084c14'), primary_key=True, serialize=False),
        ),
    ]
