# Generated by Django 3.2.5 on 2021-09-24 20:39

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('data', '0018_auto_20210814_1837'),
    ]

    operations = [
        migrations.CreateModel(
            name='PiPicoACAnalogIo',
            fields=[
                ('hardwareio_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='data.hardwareio')),
                ('pin', models.IntegerField()),
            ],
            options={
                'abstract': False,
                'base_manager_name': 'objects',
            },
            bases=('data.hardwareio',),
        ),
        migrations.AlterField(
            model_name='hardwareio',
            name='type',
            field=models.CharField(choices=[('SPI', 'SPI'), ('Serial', 'Serial'), ('PWM', 'PWM'), ('I2C', 'I2C'), ('Device File', 'Device File'), ('MCP Channel', 'MCP Channel'), ('Pi Pico Analog', 'Pi Pico Analog'), ('Pi Pico AC Analog', 'Pi Pico AC Analog'), ('Pi GPIO', 'Pi GPIO')], default='Pi Pico Analog', max_length=25),
        ),
    ]
