# Generated by Django 3.2.5 on 2021-10-06 17:43

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('data', '0035_auto_20211106_1530'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='datatransformerconstants',
            name='type',
        ),
        migrations.AlterField(
            model_name='datatransformerconstants',
            name='id',
            field=models.UUIDField(blank=True, default=uuid.UUID('66c30765-c9f0-48b9-92a5-b58b8b72f54d'), primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='datatransformerconstants',
            name='index',
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='datatransformerconstants',
            name='value',
            field=models.FloatField(blank=True, default=0.0),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='hardwaredatatransformer',
            name='hardware',
            field=models.ManyToManyField(blank=True, to='data.Channel'),
        ),
        migrations.AlterField(
            model_name='hardwarestatsdatatransformer',
            name='id',
            field=models.UUIDField(blank=True, default=uuid.UUID('2814183a-c039-48ab-8743-846ed7e2bdf2'), primary_key=True, serialize=False),
        ),
    ]
