# Generated by Django 3.2.5 on 2021-10-06 17:45

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('data', '0037_auto_20211006_1743'),
    ]

    operations = [
        migrations.AlterField(
            model_name='datatransformerconstants',
            name='id',
            field=models.UUIDField(blank=True, default=uuid.UUID('a482c481-6574-4436-991e-926f72a41a30'), primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='hardwarestatsdatatransformer',
            name='id',
            field=models.UUIDField(blank=True, default=uuid.UUID('1444a3d3-ad06-4039-bbcf-00246699b5fb'), primary_key=True, serialize=False),
        ),
    ]
