# Generated by Django 3.2.5 on 2021-10-06 15:29

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('data', '0033_auto_20211005_2134'),
    ]

    operations = [
        migrations.CreateModel(
            name='DataTransformerTypes',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False)),
                ('type', models.CharField(blank=True, default=None, max_length=15, null=True)),
            ],
        ),
        migrations.AlterField(
            model_name='datatransformer',
            name='type',
            field=models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.DO_NOTHING, to='data.datatransformertypes'),
        ),
    ]
