{% load static %}

<!DOCTYPE html>
<html xmlns="http://www.w3.org/1999/xhtml">
<head>
<meta charset="utf-8">
<meta name="viewport" content="width=device-width, initial-scale=1.0"/>
<meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
<title>Janus WebRTC Server: Streaming Demo</title>
<script type="text/javascript" src="https://cdnjs.cloudflare.com/ajax/libs/webrtc-adapter/8.0.0/adapter.min.js" ></script>
<script type="text/javascript" src="https://cdnjs.cloudflare.com/ajax/libs/jquery/1.9.1/jquery.min.js" ></script>
<script type="text/javascript" src="https://cdnjs.cloudflare.com/ajax/libs/twitter-bootstrap/3.4.1/js/bootstrap.min.js"></script>
<script type="text/javascript" src="https://cdnjs.cloudflare.com/ajax/libs/bootbox.js/5.4.0/bootbox.min.js"></script>
<script type="text/javascript" src="https://cdnjs.cloudflare.com/ajax/libs/spin.js/2.3.2/spin.min.js"></script>
<script type="text/javascript" src="https://cdnjs.cloudflare.com/ajax/libs/toastr.js/2.1.4/toastr.min.js"></script>
<script type="text/javascript" src={% static 'js/janus.js' %} ></script>
<script type="text/javascript" src={% static 'js/streamingtest.js' %}></script>
<script>
	$(function() {
		$(".navbar-static-top").load({% static 'navbar.html' %}, function() {
			$(".navbar-static-top li.dropdown").addClass("active");
			$(".navbar-static-top a[href='streamingtest.html']").parent().addClass("active");
		});
	});
</script>
<link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/bootswatch/3.4.0/cerulean/bootstrap.min.css" type="text/css"/>
<link rel="stylesheet" href={% static 'demo.css' %} type="text/css"/>
<link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/font-awesome/4.7.0/css/font-awesome.min.css" type="text/css"/>
<link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/toastr.js/2.1.4/toastr.min.css"/>
</head>
<body>

<a href="https://github.com/meetecho/janus-gateway"><img style="position: absolute; top: 0; left: 0; border: 0; z-index: 1001;" src="https://s3.amazonaws.com/github/ribbons/forkme_left_darkblue_121621.png" alt="Fork me on GitHub"></a>

<nav class="navbar navbar-default navbar-static-top">
</nav>

<div class="container">
	<div class="row">
		<div class="col-md-12">
			<div class="page-header">
				<h1>Plugin Demo: Streaming
					<button class="btn btn-default" autocomplete="off" id="start">Start</button>
				</h1>
			</div>
			<div class="container" id="details">
				<div class="row">
					<div class="col-md-12">
						<h3>Demo details</h3>
						<p>This demo showcases the functionality provided by the Streaming plugin.
						In particular, it provides three different streaming approaches, namely:</p>
						<p><ol>
							<li>An on-demand stream originated by a file (a song, in this case):
							different users accessing this stream would receive a personal view
							of the stream itself.</li>
							<li>A pseudo-live stream, still originated by a file (an audio recording
							of a radio running commentary): different users accessing this stream
							would all receive the same, shared view of the stream.</li>
							<li>A live stream, originated by a gstreamer script: as for the pseudo-live
							stream, different users will get the same feed.</li>
						</ol></p>
						<p>You can try them all within the same session: just choose the stream
						you're interested in and press the <code>Watch</code> button to start
						the playout. Stopping it will allow you to switch to a different one.</p>
						<p>Press the <code>Start</code> button above to launch the demo.</p>
					</div>
				</div>
			</div>
			<div class="container hide" id="streams">
				<div class="col-md-6">
					<div class="row">
						<div class="panel panel-default">
							<div class="panel-heading">
								<h3 class="panel-title">Streams <i id="update-streams" class="fa fa-refresh" title="Update list of streams" style="cursor: pointer;"></i></h3>
							</div>
							<div class="panel-body" id="list">
								<div class="btn-group btn-group-sm">
									<button class="btn btn-primary" autocomplete="off" id="watch">Watch or Listen</button>
									<div class="btn-group btn-group-sm">
										<button autocomplete="off" id="streamset" class="btn btn-default dropdown-toggle" data-toggle="dropdown">
											Streams list<span class="caret"></span>
										</button>
										<ul id="streamslist" class="dropdown-menu" role="menu">
										</ul>
									</div>
								</div>
							</div>
						</div>
					</div>
					<div class="row hide" id="info">
						<div class="panel panel-default">
							<div class="panel-heading">
								<h3 class="panel-title"><i class="fa fa-info-circle"></i> Metadata</h3>
							</div>
							<div class="panel-body">
								<pre id="metadata" style="word-break: break-word;"></pre>
							</div>
						</div>
					</div>
				</div>
				<div class="col-md-6">
					<div class="panel panel-default">
						<div class="panel-heading">
							<h3 class="panel-title">Stream
								<span class="label label-info hide" id="status"></span>
								<span class="label label-primary hide" id="curres"></span>
								<span class="label label-info hide" id="curbitrate"></span>
							</h3>
						</div>
						<div class="panel-body" id="stream"></div>
					</div>
				</div>
			</div>
		</div>
	</div>

	<hr>
	<div class="footer">
	</div>
</div>

</body>
</html>
