# Generated by Django 3.2.4 on 2021-07-12 23:29

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('data', '0011_auto_20210712_2041'),
    ]

    operations = [
        migrations.AddField(
            model_name='hardwareio',
            name='type',
            field=models.CharField(choices=[('SPI', 'SPI'), ('Serial', 'Serial'), ('PWM', 'PWM'), ('I2C', 'I2C'), ('Device File', 'Device File'), ('MCP Channel', 'MCP Channel'), ('Pi Pico Analog', 'Pi Pico Analog'), ('Pi GPIO', 'Pi GPIO')], default='Pi Pico Analog', max_length=25),
        ),
    ]
